import { Address, PublicKey } from '../../crypto';
import { DDOAttribute } from '../../transaction/ddo';
import { Transaction } from '../../transaction/transaction';
/**
 * Address of DNA ID contract
 */
export declare const DNAID_CONTRACT = "0000000000000000000000000000000000000003";
/**
 * Registers Identity.
 *
 * GAS calculation: gasLimit * gasPrice is equal to the amount of gas consumed.
 *
 * @param dnaid User's DNA ID
 * @param publicKey Public key
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 */
export declare function buildRegisterDNAidTx(dnaid: string, publicKey: PublicKey, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Registers Identity with initial attributes.
 *
 * @param dnaid User's DNA ID
 * @param attributes Array of DDOAttributes
 * @param publicKey User's public key
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 */
export declare function buildRegIdWithAttributes(dnaid: string, attributes: DDOAttribute[], publicKey: PublicKey, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Adds attributes to DNA ID.
 *
 * @param dnaid User's DNA ID
 * @param attributes Array of DDOAttributes
 * @param publicKey User's public key
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 */
export declare function buildAddAttributeTx(dnaid: string, attributes: DDOAttribute[], publicKey: PublicKey, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Removes attribute from DNA ID.
 *
 * @param dnaid User's DNA ID
 * @param key Key of attribute to remove
 * @param publicKey User's public key
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 *
 */
export declare function buildRemoveAttributeTx(dnaid: string, key: string, publicKey: PublicKey, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Queries attributes attached to DNA ID.
 *
 * @param dnaid User's DNA ID
 */
export declare function buildGetAttributesTx(dnaid: string): Transaction;
/**
 * Queries Description Object of DNA ID(DDO).
 *
 * @param dnaid User's DNA ID
 */
export declare function buildGetDDOTx(dnaid: string): Transaction;
/**
 * Adds a new public key to DNA ID.
 *
 * @param dnaid User's DNA ID
 * @param newPk New public key to be added
 * @param userKey User's public key or address
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 */
export declare function buildAddControlKeyTx(dnaid: string, newPk: PublicKey, userKey: PublicKey | Address, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Revokes a public key from DNA ID.
 *
 * @param dnaid User's DNA ID
 * @param pk2Remove Public key to be removed
 * @param sender User's public key or address
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 */
export declare function buildRemoveControlKeyTx(dnaid: string, pk2Remove: PublicKey, sender: PublicKey | Address, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Queries public keys attached to DNA ID.
 *
 * @param dnaid User's DNA ID
 */
export declare function buildGetPublicKeysTx(dnaid: string): Transaction;
/**
 * Adds recovery address to DNA ID.
 *
 * @param dnaid User's DNA ID
 * @param recovery Recovery address, must have not be set
 * @param publicKey User's public key, must be user's existing public key
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 */
export declare function buildAddRecoveryTx(dnaid: string, recovery: Address, publicKey: PublicKey, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Changes recovery address of DNA ID.
 *
 * This contract call must be initiated by the original recovery address.
 *
 * @param dnaid user's DNA ID
 * @param newrecovery New recovery address
 * @param oldrecovery Original recoevery address
 * @param gasPrice Gas price
 * @param gasLimit Gas limit
 * @param payer Payer
 */
export declare function buildChangeRecoveryTx(dnaid: string, newrecovery: Address, oldrecovery: Address, gasPrice: string, gasLimit: string, payer?: Address): Transaction;
/**
 * Queries the state of the public key associated with DNA ID.
 *
 * @param dnaid user's DNA ID
 * @param pkId User's public key Id
 */
export declare function buildGetPublicKeyStateTx(dnaid: string, pkId: number): Transaction;
